from ioos_qc.config_creator import (
    QC_CONFIG_CREATOR_SCHEMA,
    VARIABLE_CONFIG_SCHEMA,
    CreatorConfig,
    QcConfigCreator,
    QcVariableConfig,
)

if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "creator_config_file",
        type=str,
        help="path to creator config file (e.g. creator_config.json)",
    )
    parser.add_argument(
        "variable_config_file",
        type=str,
        help="path to variable config file (e.g. air_config.json)",
    )
    args = parser.parse_args()

    creator_config = CreatorConfig(
        args.creator_config_file,
        QC_CONFIG_CREATOR_SCHEMA,
    )
    variable_config = QcVariableConfig(
        args.variable_config_file,
        VARIABLE_CONFIG_SCHEMA,
    )
    qc = QcConfigCreator(creator_config)
    qc.create_config(variable_config)



"""Deprecated module. Consider using ARGO instead."""

import warnings

from ioos_qc import argo


def pressure_check(inp):
    warnings.warn(
        "gliders.pressure_check has been replaced by argo.pressure_increasing_test",
        DeprecationWarning,
        stacklevel=2,
    )
    return argo.pressure_increasing_test(inp)



"""Utilities."""

from __future__ import annotations

import io
import json
import logging
import re
from collections import OrderedDict
from collections.abc import Mapping
from datetime import date, datetime
from numbers import Real
from pathlib import Path
from typing import Any

import geojson
import numpy as np
import pandas as pd
import xarray as xr
from geographiclib.geodesic import Geodesic
from ruamel.yaml import YAML

N = Real
L = logging.getLogger(__name__)


def add_flag_metadata(**kwargs):
    """Add flag metadata."""

    def wrapper(func: callable):
        for k, v in kwargs.items():
            setattr(func, k, v)
        return func

    return wrapper


def openf(p, **kwargs):
    """Allow one-line-lambdas to read file contents."""
    with Path(p).open(**kwargs) as f:
        return f.read()


def load_config_from_xarray(source):
    """Load an xarray dataset as a config dict."""
    to_close = False
    if not isinstance(source, xr.Dataset):
        source = xr.open_dataset(source, decode_cf=False)
        to_close = True

    # If a global attribute exists, load as a YAML or JSON string and
    # ignore any config at the variable level
    if "ioos_qc_config" in source.attrs:
        L.info("Using global attribute ioos_qc_config for QC config")
        return load_config_as_dict(source.attrs["ioos_qc_config"])

    # Iterate over variables and construct a config object
    y = OrderedDict()
    qc_dataset = source.filter_by_attrs(
        ioos_qc_module=lambda x: x is not None,
        ioos_qc_test=lambda x: x is not None,
        ioos_qc_config=lambda x: x is not None,
        ioos_qc_target=lambda x: x is not None,
    )

    for dv in qc_dataset.data_vars:
        if dv in qc_dataset.dims:
            continue

        vobj = qc_dataset[dv]

        try:
            # Because a data variables can have more than one check
            # associated with it we need to merge any existing configs
            # for this variable
            newdict = OrderedDict(
                {
                    vobj.ioos_qc_module: OrderedDict(
                        {
                            vobj.ioos_qc_test: OrderedDict(
                                json.loads(vobj.ioos_qc_config),
                            ),
                        },
                    ),
                },
            )
            merged = dict_update(
                y.get(vobj.ioos_qc_target, {}),
                newdict,
            )
            y[vobj.ioos_qc_target] = merged
        except BaseException:  # noqa: BLE001
            L.error(f"Could not pull QC config from {vobj.name}, skipping")
            continue

    # If we opened this xarray dataset from a file we should close it
    if to_close is True:
        source.close()

    return y


def load_config_as_dict(
    source: str | dict | OrderedDict | Path | io.StringIO,
) -> OrderedDict:
    """Load an object as a config dict. The source can be a dict, OrderedDict,
    YAML string, JSON string, a StringIO, or a file path to a valid YAML or
    JSON file.

    """
    yaml = YAML(typ="safe")
    if isinstance(source, OrderedDict):
        return source
    if isinstance(source, dict):
        return OrderedDict(source)
    if isinstance(source, xr.Dataset):
        return load_config_from_xarray(source)
    if isinstance(source, (str, Path)):
        source = str(source)

        # Try to load as YAML, then JSON, then file path
        load_funcs = [
            lambda x: OrderedDict(yaml.load(x)),
            lambda x: OrderedDict(json.loads(x)),
            lambda x: load_config_from_xarray(x),
            lambda x: OrderedDict(yaml.load(openf(x))),
            lambda x: OrderedDict(json.loads(openf(x))),
        ]
        for lf in load_funcs:
            try:
                return lf(source)
            except BaseException:  # noqa: S112, PERF203, BLE001
                continue

    elif isinstance(source, io.StringIO):
        # Try to load as YAML, then JSON, then file path
        load_funcs = [
            lambda x: OrderedDict(yaml.load(x)),
            lambda x: OrderedDict(json.load(x)),
        ]
        for lf in load_funcs:
            try:
                return lf(source.getvalue())
            except BaseException:  # noqa: S112, PERF203, BLE001
                continue

    msg = "Config source is not valid!"
    raise ValueError(msg)


def isfixedlength(
    lst: list | tuple,
    length: int,
) -> bool:
    """Check if a list has the correct length."""
    if not isinstance(lst, (list, tuple)):
        msg = f"Required: list/tuple, Got: {type(lst)}"
        raise TypeError(msg)

    if len(lst) != length:
        msg = f"Incorrect list/tuple length for {lst}. Required: {length}, Got: {len(lst)}"
        raise ValueError(
            msg,
        )

    return True


def isnan(v: Any) -> bool:
    """Return True if a value is NaN."""
    return v is None or np.isnan(v).any() or v is np.ma.masked


def mapdates(dates):
    """Map dates objects to datetime64[ns]."""
    if hasattr(dates, "dtype") and hasattr(dates.dtype, "tz"):
        # pandas time objects with a datetime component, remove the timezone
        return dates.dt.tz_localize(None).astype("datetime64[ns]").to_numpy()
    if hasattr(dates, "dtype") and hasattr(dates, "to_numpy"):
        # pandas time objects without a datetime component
        return dates.to_numpy().astype("datetime64[ns]")
    if hasattr(dates, "dtype") and np.issubdtype(dates.dtype, np.datetime64):
        # numpy datetime objects
        return dates.astype("datetime64[ns]")
    try:
        # Finally try unix epoch seconds
        return (
            pd.to_datetime(dates, unit="s")
            .to_numpy()
            .astype(
                "datetime64[ns]",
            )
        )
    except Exception:  # noqa: BLE001
        # strings work here but we don't advertise that
        return np.array(dates, dtype="datetime64[ns]")


def check_timestamps(
    times: np.ndarray,
    max_time_interval: N = None,
) -> bool:
    """Sanity checks for timestamp arrays.

    Checks that the times supplied are in monotonically increasing
    chronological order, and optionally that time intervals between
    measurements do not exceed a value `max_time_interval`.  Note that this is
    not a QARTOD test, but rather a utility test to make sure times are in the
    proper order and optionally do not have large gaps prior to processing the
    data.

    Parameters
    ----------
    times
        Input array of timestamps
    max_time_interval
        The interval between values should not exceed this value. [optional]

    """
    time_diff = np.diff(times)
    sort_diff = np.diff(sorted(times))
    # Check if there are differences between sorted and unsorted, and then
    # see if if there are any duplicate times.  Then check that none of the
    # diffs exceeds the sorted time.
    zero = np.array(0, dtype=time_diff.dtype)
    return not (
        not np.array_equal(time_diff, sort_diff)
        or np.any(sort_diff == zero)
        or (max_time_interval is not None and np.any(sort_diff > max_time_interval))
    )


def dict_update(d: Mapping, u: Mapping) -> Mapping:
    """Update value of a nested dictionary of varying depth."""
    # http://stackoverflow.com/a/3233356
    for k, v in u.items():
        if isinstance(d, Mapping):
            if isinstance(v, Mapping):
                r = dict_update(d.get(k, {}), v)
                d[k] = r
            else:
                d[k] = v
        else:
            d = {k: v}
    return d


def dict_depth(d):
    """Get the depth of a dict."""
    # https://stackoverflow.com/a/23499101
    if isinstance(d, dict):
        return 1 + (max(map(dict_depth, d.values())) if d else 0)
    return 0


def cf_safe_name(name: str) -> str:
    """Make CF safe names."""
    if isinstance(name, str):
        if re.match("^[0-9_]", name):
            # Add a letter to the front
            name = f"v_{name}"
        return re.sub(r"[^_a-zA-Z0-9]", "_", name)

    msg = f'Could not convert "{name}" to a safe name'
    raise ValueError(msg)


class GeoNumpyDateEncoder(geojson.GeoJSONEncoder):
    """Encode numpy dates for geospatial formats."""

    def default(self, obj: Any) -> Any:
        """If input object is an ndarray it will be converted into a list."""
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        if isinstance(obj, np.generic):
            return obj.item()
        if isinstance(obj, (datetime, date)):
            return obj.isoformat()
        if np.isnan(obj):
            return None

        return geojson.factory.GeoJSON.to_instance(obj)


def great_circle_distance(lat_arr, lon_arr):
    """Compute great circle distances."""

    def gc(y1, x1, y2, x2):
        return Geodesic.WGS84.Inverse(y1, x1, y2, x2)["s12"]

    dist = np.ma.zeros(lon_arr.size, dtype=np.float64)
    dv = np.vectorize(gc)
    dist[1:] = dv(lat_arr[:-1], lon_arr[:-1], lat_arr[1:], lon_arr[1:])
    return dist



from __future__ import annotations

import inspect
import json
import logging
from importlib import import_module
from pathlib import Path
from typing import TYPE_CHECKING

import h5netcdf.legacyapi as nc4
import numpy as np
import pandas as pd

from ioos_qc.qartod import aggregate
from ioos_qc.results import CollectedResult, collect_results
from ioos_qc.utils import GeoNumpyDateEncoder, cf_safe_name

if TYPE_CHECKING:
    from ioos_qc.config import Config

L = logging.getLogger(__name__)


def column_from_collected_result(cr):
    stream_label = f"{cr.stream_id}." if cr.stream_id else ""
    package_label = f"{cr.package}." if cr.package else ""
    test_label = f"{cr.test}" if cr.test else ""
    return cf_safe_name(f"{stream_label}{package_label}{test_label}")


class BaseStore:
    def save(self, *args, **kwargs) -> None:
        """Serialize results to a store. This could save a file or publish messages."""

    @property
    def stream_ids(self) -> list[str]:
        """A list of stream_ids to save to the store."""


class PandasStore(BaseStore):
    """Store results in a dataframe."""

    def __init__(self, results, axes: dict | None = None) -> None:
        # OK, time to evaluate the actual tests now that we need the results
        self.results = list(results)
        self.collected_results = collect_results(self.results, how="list")
        self._stream_ids = [cr.stream_id for cr in self.collected_results]
        self.axes = axes or {
            "t": "time",
            "z": "z",
            "y": "lat",
            "x": "lon",
        }

    @property
    def stream_ids(self) -> list[str]:
        return self._stream_ids

    def compute_aggregate(self, name="rollup") -> None:
        """Internally compute the total aggregate and add it to the results."""
        agg = CollectedResult(
            stream_id="",
            package="qartod",
            test=name,
            function=aggregate,
            results=aggregate(self.collected_results),
        )
        self.collected_results.append(agg)

    def save(
        self,
        *,
        write_data: bool = False,
        write_axes: bool = True,
        include: list | None = None,
        exclude: list | None = None,
    ) -> pd.DataFrame:
        save_df = pd.DataFrame()

        for cr in self.collected_results:
            # Add time axis
            if write_axes is True and self.axes["t"] not in save_df and cr.tinp is not None and cr.tinp.size != 0:
                L.info(
                    f"Adding column {self.axes['t']} from stream {cr.stream_id}",
                )
                save_df[self.axes["t"]] = cr.tinp

            # Add z axis
            if write_axes is True and self.axes["z"] not in save_df and cr.zinp is not None and cr.zinp.size != 0:
                L.info(
                    f"Adding column {self.axes['z']} from stream {cr.stream_id}",
                )
                save_df[self.axes["z"]] = cr.zinp

            # Add x axis
            if write_axes is True and self.axes["x"] not in save_df and cr.lon is not None and cr.lon.size != 0:
                L.info(
                    f"Adding column {self.axes['x']} from stream {cr.stream_id}",
                )
                save_df[self.axes["x"]] = cr.lon

            # Add x axis
            if write_axes is True and self.axes["y"] not in save_df and cr.lat is not None and cr.lat.size != 0:
                L.info(
                    f"Adding column {self.axes['y']} from stream {cr.stream_id}",
                )
                save_df[self.axes["y"]] = cr.lat

            # Inclusion list, skip everything not defined
            if include is not None and (cr.function not in include and cr.stream_id not in include and cr.test not in include):
                continue

            # Exclusion list, skip everything defined
            if exclude is not None and (cr.function in exclude or cr.stream_id in exclude or cr.test in cr.test in include):
                continue

            # Add data column
            if write_data and cr.stream_id not in save_df and cr.stream_id:
                L.info(f"Adding column {cr.stream_id}")
                save_df[cr.stream_id] = cr.data

            # Add QC results column
            # Aggregate will have None stream_id, so allow it to be that way!
            column_name = column_from_collected_result(cr)
            if column_name not in save_df:
                save_df[column_name] = cr.results
            else:
                L.warning(
                    f"Found duplicate QC results column: {column_name}, skipping.",
                )

        return save_df


class CFNetCDFStore(BaseStore):
    def __init__(self, results, axes=None) -> None:
        # OK, time to evaluate the actual tests now that we need the results
        self.results = list(results)
        self.collected_results = collect_results(self.results, how="list")
        self._stream_ids = [cr.stream_id for cr in self.collected_results]
        self.axes = axes or {
            "t": "time",
            "z": "z",
            "y": "lat",
            "x": "lon",
        }

    @property
    def stream_ids(self) -> list[str]:
        return self._stream_ids

    def save(  # noqa: PLR0913, C901
        self,
        path_or_ncd,
        dsg,
        *,
        config: Config,
        dsg_kwargs: dict | None = None,
        write_data: bool = False,
        include: list | None = None,
        exclude: list | None = None,
        compute_aggregate: bool = False,
    ):
        if dsg_kwargs is None:
            dsg_kwargs = {}
        ps = PandasStore(self.results, self.axes)
        if compute_aggregate is True:
            ps.compute_aggregate(name="qc_rollup")

        data_df = ps.save(write_data=write_data, include=include, exclude=exclude)

        # Write a new file
        attrs = {}
        for cr in ps.collected_results:
            column_name = column_from_collected_result(cr)

            # Set the ancillary variables
            if cr.stream_id not in attrs:
                attrs[cr.stream_id] = {
                    "ancillary_variables": column_name,
                }
            else:
                # Update the source ancillary_variables
                existing = getattr(
                    attrs[cr.stream_id],
                    "ancillary_variables",
                    "",
                ).split(" ")
                existing += [column_name]
                attrs[cr.stream_id] = " ".join(list(set(existing))).strip()

            # determine standard name and long name. These should be defined on each test function
            # https://github.com/cf-convention/cf-conventions/issues/216
            standard_name = getattr(
                cr.function,
                "standard_name",
                "quality_flag",
            )
            long_name = getattr(cr.function, "long_name", "Quality Flag")

            # Get flags from module attribute called FLAGS
            flags = inspect.getmodule(cr.function).FLAGS
            varflagnames = [d for d in flags.__dict__ if not d.startswith("__")]
            varflagvalues = [getattr(flags, d) for d in varflagnames]

            # Set QC variable attributes
            if column_name not in attrs:
                attrs[column_name] = {
                    "standard_name": standard_name,
                    "long_name": long_name,
                    "flag_values": np.byte(varflagvalues),
                    "flag_meanings": " ".join(varflagnames),
                    "valid_min": np.byte(min(varflagvalues)),
                    "valid_max": np.byte(max(varflagvalues)),
                    "ioos_qc_module": cr.package,
                    "ioos_qc_test": cr.test,
                    "ioos_qc_target": cr.stream_id,
                }
                # If there is only one context we can write variable specific configs
                # We can't do this across different contexts and this would repeat the regions
                # and windows for each variable even if they are equal. This needs another look.
                if len(config.contexts) == 1:
                    calls = config.calls_by_stream_id(cr.stream_id)

                    calls = [c for c in calls if c.module == cr.package and c.method == cr.test]
                    if not calls:
                        # No stream_id found!
                        continue

                    # Use the first call of this stream_id. There will be only 1 because there
                    # is only one context with one matching package and method
                    call = calls[0]
                    if call.region:
                        attrs[column_name]["ioos_qc_region"] = json.dumps(
                            call.region,
                            cls=GeoNumpyDateEncoder,
                            allow_nan=False,
                            ignore_nan=True,
                        )
                    if call.window.starting or call.window.ending:
                        attrs[column_name]["ioos_qc_window"] = json.dumps(
                            call.window,
                            cls=GeoNumpyDateEncoder,
                            allow_nan=False,
                            ignore_nan=True,
                        )

                    qc_varconfig = json.dumps(
                        call.kwargs,
                        cls=GeoNumpyDateEncoder,
                        allow_nan=False,
                        ignore_nan=True,
                    )
                    attrs[column_name]["ioos_qc_config"] = qc_varconfig

        if len(config.contexts) > 1:
            # We represent the config as one global config JSON object
            attrs["ioos_qc_config"] = json.dumps(
                config.config,
                cls=GeoNumpyDateEncoder,
                allow_nan=False,
                ignore_nan=True,
            )

        dsg_kwargs = {
            **dsg_kwargs,
            "attributes": attrs,
        }

        # pocean requires these default columns, which should be removed as a requirement
        # in pocean.
        data_df["station"] = 0
        data_df["trajectory"] = 0
        data_df["profile"] = 0
        if "z" not in data_df:
            data_df["z"] = 0
        return dsg.from_dataframe(
            data_df,
            path_or_ncd,
            axes=self.axes,
            **dsg_kwargs,
        )


class NetcdfStore:
    def save(self, path_or_ncd, config, results):  # noqa: PLR0912, PLR0915, C901
        """Updates the given netcdf with test configuration and results.
        If there is already a variable for a given test, it will update that variable with the latest results.
        Otherwise, it will create a new variable.

        :param path_or_ncd: path or netcdf4 Dataset in which to store results
        :param results: output of run()
        """
        try:
            ncd = None
            should_close = True
            if isinstance(path_or_ncd, (str, Path)):
                ncd = nc4.Dataset(str(path_or_ncd), "a")
            elif isinstance(path_or_ncd, nc4.Dataset):
                ncd = path_or_ncd
                should_close = False
            else:
                return ValueError("Input is not a valid file path or Dataset")

            for vname, qcobj in results.items():
                if vname not in ncd.variables:
                    L.warning(f"{vname} not found in the Dataset, skipping")
                    continue

                source_var = ncd.variables[vname]
                # Keep track of the test names so we can add to the source's
                # ancillary_variables at the end
                qcvar_names = []

                for modu, tests in qcobj.items():
                    try:
                        testpackage = import_module(f"ioos_qc.{modu}")
                    except ImportError:
                        L.error(
                            f'No ioos_qc test package "{modu}" was found, skipping.',
                        )
                        continue

                    for testname, testresults in tests.items():
                        # Try to find a qc variable that matches this config
                        qcvars = ncd.get_variables_by_attributes(
                            ioos_qc_module=modu,
                            ioos_qc_test=testname,
                            ioos_qc_target=vname,
                        )
                        if not qcvars:
                            qcvarname = cf_safe_name(
                                vname + "." + modu + "." + testname,
                            )
                        else:
                            if len(qcvars) > 1:
                                names = [v.name for v in qcvars]
                                L.warning(
                                    f"Found more than one QC variable match: {names}",
                                )
                            # Use the last one found
                            qcvarname = qcvars[-1].name

                        # Get flags from module attribute called FLAGS
                        flags = testpackage.FLAGS
                        varflagnames = [d for d in flags.__dict__ if not d.startswith("__")]
                        varflagvalues = [getattr(flags, d) for d in varflagnames]

                        v = ncd.createVariable(qcvarname, np.byte, source_var.dimensions) if qcvarname not in ncd.variables else ncd[qcvarname]

                        # determine standard name
                        # https://github.com/cf-convention/cf-conventions/issues/216
                        try:
                            testfn = getattr(testpackage, testname)
                            standard_name = testfn.standard_name
                            long_name = testfn.long_name
                        except AttributeError:
                            standard_name = "quality_flag"
                            long_name = "Quality Flag"

                        # write to netcdf
                        v[:] = testresults
                        v.setncattr("standard_name", standard_name)
                        v.setncattr("long_name", long_name)
                        v.setncattr("flag_values", np.byte(varflagvalues))
                        v.setncattr("flag_meanings", " ".join(varflagnames))
                        v.setncattr("valid_min", np.byte(min(varflagvalues)))
                        v.setncattr("valid_max", np.byte(max(varflagvalues)))
                        v.setncattr("ioos_qc_module", modu)
                        v.setncattr("ioos_qc_test", testname)
                        v.setncattr("ioos_qc_target", vname)
                        # If there is only one context we can write variable specific configs
                        if len(config.contexts) == 1:
                            varconfig = config.contexts[0].streams[vname].config[modu][testname]
                            varconfig = json.dumps(
                                varconfig,
                                cls=GeoNumpyDateEncoder,
                                allow_nan=False,
                                ignore_nan=True,
                            )
                            v.setncattr("ioos_qc_config", varconfig)
                            v.setncattr(
                                "ioos_qc_region",
                                json.dumps(
                                    config.contexts[0].region,
                                    cls=GeoNumpyDateEncoder,
                                    allow_nan=False,
                                    ignore_nan=True,
                                ),
                            )
                            v.setncattr(
                                "ioos_qc_window",
                                json.dumps(
                                    config.contexts[0].window,
                                    cls=GeoNumpyDateEncoder,
                                    allow_nan=False,
                                    ignore_nan=True,
                                ),
                            )

                # Update the source ancillary_variables
                existing = getattr(
                    source_var,
                    "ancillary_variables",
                    "",
                ).split(" ")
                if qcvar_names:
                    existing += qcvar_names
                source_var.ancillary_variables = " ".join(
                    list(set(existing)),
                ).strip()

            if len(config.contexts) > 1:
                # We can't represent these at the variable level, so make one global config
                ncd.setncattr(
                    "ioos_qc_config",
                    json.dumps(
                        config.config,
                        cls=GeoNumpyDateEncoder,
                        allow_nan=False,
                        ignore_nan=True,
                    ),
                )

        finally:
            if ncd and should_close is True:
                ncd.close()



import logging

import numpy as np
from bokeh import plotting
from bokeh.layouts import gridplot

L = logging.getLogger(__name__)
flag_pass, flag_notrun, flag_suspect, flag_fail = 1, 2, 3, 4


def bokeh_plot(data, var_name, results, title, module, test_name):  # noqa: PLR0913
    plot = bokeh_plot_var(data, var_name, results, title, module, test_name)
    return gridplot([[plot]], sizing_mode="fixed")


def bokeh_plot_var(time, data, results, title, module, test_name):  # noqa: PLR0913
    """Method to plot QC results using Bokeh."""
    if module not in results or test_name not in results[module]:
        L.warning(f"No results for test {module}.{test_name} found")
        return None

    qc_test = results[module][test_name]

    qc_pass = np.ma.masked_where(qc_test != flag_pass, data)
    qc_notrun = np.ma.masked_where(qc_test != flag_notrun, data)
    qc_suspect = np.ma.masked_where(qc_test != flag_suspect, data)
    qc_fail = np.ma.masked_where(qc_test != flag_fail, data)

    p1 = plotting.figure(
        x_axis_type="datetime",
        title=test_name + " : " + title,
    )
    p1.grid.grid_line_alpha = 0.3
    p1.xaxis.axis_label = "Time"
    p1.yaxis.axis_label = "Data"

    p1.line(time, data, legend_label="data", color="#A6CEE3")
    p1.circle(
        time,
        qc_notrun,
        size=2,
        legend_label="qc not run",
        color="gray",
        alpha=0.2,
    )
    p1.circle(
        time,
        qc_pass,
        size=4,
        legend_label="qc pass",
        color="green",
        alpha=0.5,
    )
    p1.circle(
        time,
        qc_suspect,
        size=4,
        legend_label="qc suspect",
        color="orange",
        alpha=0.7,
    )
    p1.circle(
        time,
        qc_fail,
        size=6,
        legend_label="qc fail",
        color="red",
        alpha=1.0,
    )
    p1.circle(
        time,
        qc_notrun,
        size=6,
        legend_label="qc not eval",
        color="gray",
        alpha=1.0,
    )

    return p1


def bokeh_multi_plot(stream, results, title, **kwargs):
    kwargs = {
        "merge_tools": True,
        "toolbar_location": "below",
        "sizing_mode": "scale_width",
        "plot_width": 600,
        "plot_height": 200,
        "ncols": 2,
        **kwargs,
    }

    plots = list(bokeh_multi_var(stream, results, title))
    return gridplot(plots, **kwargs)


def bokeh_multi_var(stream, results, title):
    for vname, qcobj in results.items():
        for modu, tests in qcobj.items():
            for testname in tests:
                plt = bokeh_plot_var(
                    stream.time(),
                    stream.data(vname),
                    vname,
                    qcobj,
                    title,
                    modu,
                    testname,
                )
                yield plt


def bokeh_plot_collected_results(results, **kwargs):
    kwargs = {
        "merge_tools": True,
        "toolbar_location": "below",
        "sizing_mode": "scale_width",
        "plot_width": 600,
        "plot_height": 200,
        "ncols": 2,
        **kwargs,
    }

    plots = [bokeh_plot_collected_result(r) for r in results if r.data.any() and r.results.any()]
    return gridplot(plots, **kwargs)


def bokeh_plot_collected_result(cr):
    title = f"{cr.stream_id}: {cr.package}.{cr.test}"
    p1 = plotting.figure(x_axis_type="datetime", title=title)
    p1.grid.grid_line_alpha = 0.3
    p1.xaxis.axis_label = "Time"
    p1.yaxis.axis_label = "Data"

    qc_pass = np.ma.masked_where(cr.results != flag_pass, cr.data)
    qc_notrun = np.ma.masked_where(cr.results != flag_notrun, cr.data)
    qc_suspect = np.ma.masked_where(cr.results != flag_suspect, cr.data)
    qc_fail = np.ma.masked_where(cr.results != flag_fail, cr.data)

    p1.line(cr.tinp, cr.data, legend_label="data", color="#A6CEE3")
    p1.circle(
        cr.tinp,
        qc_notrun,
        size=3,
        legend_label="qc not run",
        color="gray",
        alpha=0.2,
    )
    p1.circle(
        cr.tinp,
        qc_pass,
        size=4,
        legend_label="qc pass",
        color="green",
        alpha=0.5,
    )
    p1.circle(
        cr.tinp,
        qc_suspect,
        size=4,
        legend_label="qc suspect",
        color="orange",
        alpha=0.7,
    )
    p1.circle(
        cr.tinp,
        qc_fail,
        size=6,
        legend_label="qc fail",
        color="red",
        alpha=1.0,
    )
    p1.circle(
        cr.tinp,
        qc_notrun,
        size=3,
        legend_label="qc not eval",
        color="gray",
        alpha=1.0,
    )

    return p1



"""Tests based on the IOOS QARTOD manuals."""

from __future__ import annotations

import logging
import warnings
from collections import namedtuple
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Sequence
    from numbers import Real

import numpy as np
import pandas as pd

try:
    from numba.core.errors import NumbaTypeError
except ImportError:
    NumbaTypeError = TypeError

from ioos_qc.utils import (
    add_flag_metadata,
    great_circle_distance,
    isfixedlength,
    isnan,
    mapdates,
)

L = logging.getLogger(__name__)


class QartodFlags:
    """Primary flags for QARTOD."""

    GOOD = 1
    UNKNOWN = 2
    SUSPECT = 3
    FAIL = 4
    MISSING = 9


FLAGS = QartodFlags  # Default name for all check modules
NOTEVAL_VALUE = QartodFlags.UNKNOWN

WEEK_PERIODS = [
    "week",
    "weekofyear",
]

span = namedtuple("Span", "minv maxv")  # noqa: PYI024


@add_flag_metadata(
    standard_name="aggregate_quality_flag",
    long_name="Aggregate Quality Flag",
    aggregate=True,
)
def aggregate(results: list) -> np.ma.MaskedArray:
    """Runs qartod_compare against all other qartod tests in results."""
    all_tests = [r.results for r in results]
    return qartod_compare(all_tests)


def qartod_compare(
    vectors: Sequence[Sequence[Real]],
) -> np.ma.MaskedArray:
    """Aggregates an array of flags by precedence into a single array.

    Parameters
    ----------
    vectors
        An array of uniform length arrays representing individual flags

    Returns
    -------
    flag_arr
        A masked array of aggregated flag data.

    """
    shapes = [v.shape[0] for v in vectors]
    if not all(v.ndim == 1 for v in vectors) or not all(s == shapes[0] for s in shapes):
        msg = f"Vectors are not the same size.\n{vectors=}"
        raise ValueError(msg)

    result = np.ma.empty(shapes[0])
    result.fill(QartodFlags.MISSING)

    priorities = [
        QartodFlags.MISSING,
        QartodFlags.UNKNOWN,
        QartodFlags.GOOD,
        QartodFlags.SUSPECT,
        QartodFlags.FAIL,
    ]
    # For each of the priorities in order, set the resultant array to the the
    # flag where that flag exists in each of the vectors.
    for p in priorities:
        for v in vectors:
            idx = np.where(v == p)[0]
            result[idx] = p
    return result.astype("uint8")


@add_flag_metadata(
    standard_name="location_test_quality_flag",
    long_name="Location Test Quality Flag",
)
def location_test(
    lon: Sequence[Real],
    lat: Sequence[Real],
    bbox: tuple[Real, Real, Real, Real] = (-180, -90, 180, 90),
    range_max: Real | None = None,
) -> np.ma.core.MaskedArray:
    """Checks that a location is within reasonable bounds.

    Checks that longitude and latitude are within reasonable bounds defaulting
    to lon = [-180, 180] and lat = [-90, 90].  Optionally, check for a maximum
    range parameter in great circle distance defaulting to meters which can
    also use a unit from the quantities library. Missing and masked data is
    flagged as UNKNOWN.

    Parameters
    ----------
    lon
        Longitudes as a numeric numpy array or a list of numbers.
    lat
        Latitudes as a numeric numpy array or a list of numbers.
    bbox
        A length 4 tuple expressed in (minx, miny, maxx, maxy) [optional].
    range_max
        Maximum allowed range expressed in geodesic curve distance (meters).

    Returns
    -------
    flag_arr
        A masked array of flag values equal in size to that of the input.

    """
    bboxnt = namedtuple("BBOX", "minx miny maxx maxy")  # noqa: PYI024
    if bbox is not None:
        if not isfixedlength(bbox, 4):
            msg = f"{bbox=}, expecred 4."
            raise ValueError(msg)
        bbox = bboxnt(*bbox)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        lat = np.ma.masked_invalid(np.array(lat).astype(np.float64))
        lon = np.ma.masked_invalid(np.array(lon).astype(np.float64))

    if lon.shape != lat.shape:
        msg = f"Lon ({lon.shape}) and lat ({lat.shape}) are different shapes"
        raise ValueError(
            msg,
        )

    # Save original shape
    original_shape = lon.shape
    lon = lon.flatten()
    lat = lat.flatten()

    # Start with everything as passing (1)
    flag_arr = np.ma.ones(lon.size, dtype="uint8")

    # If either lon or lat are masked we just set the flag to MISSING
    mloc = lon.mask & lat.mask
    flag_arr[mloc] = QartodFlags.MISSING

    # If there is only one masked value fail the location test
    mismatch = lon.mask != lat.mask
    flag_arr[mismatch] = QartodFlags.FAIL

    if range_max is not None and lon.size > 1:
        # Calculating the great_distance between each point
        # Flag suspect any distance over range_max
        d = great_circle_distance(lat, lon)
        flag_arr[d > range_max] = QartodFlags.SUSPECT

    # Ignore warnings when comparing NaN values even though they are masked
    # https://github.com/numpy/numpy/blob/master/doc/release/1.8.0-notes.rst#runtime-warnings-when-comparing-nan-numbers
    with np.errstate(invalid="ignore"):
        flag_arr[(lon < bbox.minx) | (lat < bbox.miny) | (lon > bbox.maxx) | (lat > bbox.maxy)] = QartodFlags.FAIL

    return flag_arr.reshape(original_shape)


@add_flag_metadata(
    standard_name="gross_range_test_quality_flag",
    long_name="Gross Range Test Quality Flag",
)
def gross_range_test(
    inp: Sequence[Real],
    fail_span: tuple[Real, Real],
    suspect_span: tuple[Real, Real] | None = None,
) -> np.ma.core.MaskedArray:
    """Checks that values are within reasonable range bounds.

    Given a 2-tuple of minimum/maximum values, flag data outside of the given
    range as FAIL data.  Optionally also flag data which falls outside of a user
    defined range as SUSPECT. Missing and masked data is flagged as UNKNOWN.

    Parameters
    ----------
    inp
        Input data as a numeric numpy array or a list of numbers.
    fail_span
        2-tuple range which to flag outside data as FAIL.
    suspect_span
        2-tuple range which to flag outside data as SUSPECT. [optional]

    Returns
    -------
    flag_arr
        A masked array of flag values equal in size to that of the input.

    """
    if not isfixedlength(fail_span, 2):
        msg = f"{fail_span=}, expected 2"
        raise ValueError(msg)
    sspan = span(*sorted(fail_span))

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        inp = np.ma.masked_invalid(np.array(inp).astype(np.float64))

    # Save original shape
    original_shape = inp.shape
    inp = inp.flatten()
    # Start with everything as passing (1)
    flag_arr = np.ma.ones(inp.size, dtype="uint8")

    # If the value is masked set the flag to MISSING
    flag_arr[inp.mask] = QartodFlags.MISSING

    if suspect_span is not None:
        if not isfixedlength(suspect_span, 2):
            msg = f"{suspect_span=}, expected 2."
            raise ValueError(msg)
        uspan = span(*sorted(suspect_span))
        if uspan.minv < sspan.minv or uspan.maxv > sspan.maxv:
            msg = f"Suspect {uspan} must fall within the Fail {sspan}"
            raise ValueError(msg)
        # Flag suspect outside of user span
        with np.errstate(invalid="ignore"):
            flag_arr[(inp < uspan.minv) | (inp > uspan.maxv)] = QartodFlags.SUSPECT

    # Flag suspect outside of sensor span
    with np.errstate(invalid="ignore"):
        flag_arr[(inp < sspan.minv) | (inp > sspan.maxv)] = QartodFlags.FAIL

    return flag_arr.reshape(original_shape)


class ClimatologyConfig:
    """Objects to hold the config for a Climatology test.

    Parameters
    ----------
    tspan
        2-tuple range.
        If period is defined, then this is a numeric range.
        If period is not defined, then its a date range.
    fspan
        (optional) 2-tuple range of valid values. This is passed in as the fail_span to the gross_range_test.
    vspan
        2-tuple range of valid values. This is passed in as the suspect_span to the gross_range test.
    zspan
        (optional) Vertical (depth) range, in meters positive down
    period
        (optional) The unit the tspan argument is in. Defaults to datetime object
        but can also be any attribute supported by a pandas Timestamp object.

        See: https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.Timestamp.html

        Options:
            * year
            * week / weekofyear
            * dayofyear
            * dayofweek
            * quarter

    """

    mem = namedtuple(  # noqa: PYI024
        "window",
        [
            "tspan",
            "fspan",
            "vspan",
            "zspan",
            "period",
        ],
    )

    def __init__(self, members=None) -> None:
        members = members or []
        self._members = members

    @property
    def members(self):
        return self._members

    def values(self, tind: pd.Timestamp, zind=None):
        """Parameters
        ----------
        tind
            Value to test for inclusion between time bounds

        """
        span = (None, None)
        for m in self._members:
            # If a period is defined, extract the attribute from the
            # pd.Timestamp object before comparison. The min and max
            # values are in this period unit already.
            # If a period isn't defined, make a new Timestamp object
            # to align with the above name 'tind_copy'
            tind_copy = getattr(tind, m.period) if m.period is not None else tind

            # If we are between times
            if tind_copy > m.tspan.minv and tind_copy <= m.tspan.maxv:
                if not isnan(zind) and not isnan(m.zspan):
                    # If we are between depths
                    if zind > m.zspan.minv and zind <= m.zspan.maxv:
                        span = m.vspan
                elif isnan(zind) and isnan(m.zspan):
                    span = m.vspan
        return span

    def add(
        self,
        tspan: tuple[Real, Real],
        vspan: tuple[Real, Real],
        fspan: tuple[Real, Real] | None = None,
        zspan: tuple[Real, Real] | None = None,
        period: str | None = None,
    ) -> None:
        if not isfixedlength(tspan, 2):
            msg = f"{tspan=}, expected 2."
            raise ValueError(msg)
        # If period is defined, tspan is a numeric
        # if it isn't defined, its a parsable date
        tspan = span(*sorted(tspan)) if period is not None else span(*sorted([pd.Timestamp(tspan[0]), pd.Timestamp(tspan[1])]))

        if not isfixedlength(vspan, 2):
            msg = f"{vspan=}, expected 2."
            raise ValueError(msg)
        vspan = span(*sorted(vspan))

        if fspan is not None:
            if not isfixedlength(fspan, 2):
                msg = f"{fspan=}, expected 2."
                raise ValueError(msg)
            fspan = span(*sorted(fspan))

        if zspan is not None:
            if not isfixedlength(zspan, 2):
                msg = f"{zspan=}, expected 2."
                raise ValueError(msg)
            zspan = span(*sorted(zspan))

        if period is not None:
            # Test to make sure this is callable on a Timestamp
            try:
                getattr(pd.Timestamp.now(), period)
            except AttributeError as err:
                msg = 'The period "{period}" is not recognized'
                raise ValueError(msg) from err

        self._members.append(
            self.mem(
                tspan,
                fspan,
                vspan,
                zspan,
                period,
            ),
        )

    def check(self, tinp, inp, zinp):
        # Start with everything as UNKNOWN (2)
        flag_arr = np.ma.empty(inp.size, dtype="uint8")
        flag_arr.fill(QartodFlags.UNKNOWN)

        # If the value is masked set the flag to MISSING
        flag_arr[inp.mask] = QartodFlags.MISSING

        # Iterate over each member and apply its spans on the input data.
        # Member spans are applied in order and any data points that fall into
        # more than one member are flagged by each one.
        for m in self._members:
            if m.period is not None:
                # If a period is defined, extract the attribute from the
                # pd.DatetimeIndex object before comparison. The min and max
                # values are in this period unit already.
                if m.period in WEEK_PERIODS:
                    # The weekofyear accessor was depreacated
                    tinp_copy = pd.Index(
                        tinp.isocalendar().week,
                        dtype="int64",
                    )
                else:
                    tinp_copy = getattr(tinp, m.period).to_series()
            else:
                # If a period isn't defined, make a new Timestamp object
                # to align with the above name 'tinp_copy'
                tinp_copy = tinp

            # If a zspan is defined but we don't have z input (zinp), skip this member
            # Note: `zinp.count()` can return `np.ma.masked` so we also check using isnan
            if not isnan(m.zspan) and (not zinp.count() or isnan(zinp.any())):
                continue

            # Indexes that align with the T
            t_idx = (tinp_copy >= m.tspan.minv) & (tinp_copy <= m.tspan.maxv)

            # Indexes that align with the Z
            if not isnan(m.zspan):
                # Only test non-masked values between the min and max.
                # Ignore warnings about comparing masked values
                with np.errstate(invalid="ignore"):
                    z_idx = (~zinp.mask) & (zinp >= m.zspan.minv) & (zinp <= m.zspan.maxv)
            else:
                # If there is no z data in the config, don't try to filter by depth!
                # Set z_idx to all True to prevent filtering
                # Must use inp.data to create masked array so that the masked value is ignored when we assign the FAIL, SUSPECT, and GOOD flags
                z_idx = np.ma.array(
                    data=~np.isnan(inp.data),
                    mask=inp.mask,
                    fill_value=999999,
                )

            # Combine the T and Z indexes
            values_idx = t_idx & z_idx

            # Failed and suspect data for this value span. Combining fail_idx or
            # suspect_idx with values_idx represents the subsets of data that should be
            # fail and suspect respectively.
            fail_idx = (inp < m.fspan.minv) | (inp > m.fspan.maxv) if not isnan(m.fspan) else np.zeros(inp.size, dtype=bool)

            suspect_idx = (inp < m.vspan.minv) | (inp > m.vspan.maxv)

            with np.errstate(invalid="ignore"):
                flag_arr[(values_idx & fail_idx)] = QartodFlags.FAIL
                flag_arr[(values_idx & ~fail_idx & suspect_idx)] = QartodFlags.SUSPECT
                flag_arr[(values_idx & ~fail_idx & ~suspect_idx)] = QartodFlags.GOOD

        return flag_arr

    @staticmethod
    def convert(config):
        # Create a ClimatologyConfig object if one was not passed in
        if isinstance(config, ClimatologyConfig):
            return config

        c = ClimatologyConfig()
        for climate_config_dict in config:
            c.add(**climate_config_dict)
        return c


@add_flag_metadata(
    standard_name="climatology_test_quality_flag",
    long_name="Climatology Test Quality Flag",
)
def climatology_test(
    config: ClimatologyConfig | Sequence[dict[str, tuple]],
    inp: Sequence[Real],
    tinp: Sequence[Real],
    zinp: Sequence[Real],
) -> np.ma.core.MaskedArray:
    """Checks that values are within reasonable range bounds and flags as SUSPECT.

    Data for which no ClimatologyConfig member exists is marked as UNKNOWN.

    Parameters
    ----------
    config
        A ClimatologyConfig object or a list of dicts containing tuples
        that can be used to create a ClimatologyConfig object. See ClimatologyConfig
        docs for more info.
    inp
        Input data as a numeric numpy array or a list of numbers.
    tinp
        Time data as a sequence of datetime objects compatible with pandas DatetimeIndex.
        This includes numpy datetime64, python datetime objects and pandas Timestamp object.
        ie. pd.DatetimeIndex([datetime.utcnow(), np.datetime64(), pd.Timestamp.now()])
        If anything else is passed in the format is assumed to be seconds since the unix epoch.
    zinp
        Z (depth) data, in meters positive down, as a numeric numpy array or a list of numbers.

    Returns
    -------
    flag_arr
        A masked array of flag values equal in size to that of the input.

    """
    # Create a ClimatologyConfig object if one was not passed in
    config = ClimatologyConfig.convert(config)

    tinp = mapdates(tinp)
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        inp = np.ma.masked_invalid(np.array(inp).astype(np.float64))
        zinp = np.ma.masked_invalid(np.array(zinp).astype(np.float64))

    # Save original shape
    original_shape = inp.shape

    # We compare using a pandas Timestamp for helper functions like
    # 'week' and 'dayofyear'. It is surprisingly hard to pull these out
    # of a plain datetime64 object.
    tinp = pd.DatetimeIndex(tinp.flatten())
    inp = inp.flatten()
    zinp = zinp.flatten()

    flag_arr = config.check(tinp, inp, zinp)
    return flag_arr.reshape(original_shape)


@add_flag_metadata(
    standard_name="spike_test_quality_flag",
    long_name="Spike Test Quality Flag",
)
def spike_test(
    inp: Sequence[Real],
    suspect_threshold: Real | None = None,
    fail_threshold: Real | None = None,
    method: str = "average",
) -> np.ma.core.MaskedArray:
    """Check for spikes by checking neighboring data against thresholds.

    Determine if there is a spike at data point n-1 by subtracting
    the midpoint of n and n-2 and taking the absolute value of this
    quantity, and checking if it exceeds a low or high threshold (default).
    Values which do not exceed either threshold are flagged GOOD,
    values which exceed the low threshold are flagged SUSPECT,
    and values which exceed the high threshold are flagged FAIL.
    Missing and masked data is flagged as UNKNOWN.

    Parameters
    ----------
    inp
        Input data as a numeric numpy array or a list of numbers.
    suspect_threshold
        The SUSPECT threshold value, in observations units.
    fail_threshold
        The SUSPECT threshold value, in observations units.
    method
        ['average'(default),'differential'] optional input to assign the method used to detect spikes.
            * "average": Determine if there is a spike at data point n-1 by subtracting
                        the midpoint of n and n-2 and taking the absolute value of this
                        quantity, and checking if it exceeds a low or high threshold.
            * "differential": Determine if there is a spike at data point n by calculating the difference
                              between n and n-1 and n+1 and n variation. To considered, (n - n-1)*(n+1 - n) should
                              be smaller than zero (in opposite direction).

    Returns
    -------
    flag_arr
        A masked array of flag values equal in size to that of the input.

    """
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        inp = np.ma.masked_invalid(np.array(inp).astype(np.float64))

    # Save original shape
    original_shape = inp.shape
    inp = inp.flatten()

    # Apply different method
    if method == "average":
        # Calculate the average of n-2 and n
        ref = np.ma.zeros(inp.size, dtype=np.float64)
        ref[1:-1] = (inp[0:-2] + inp[2:]) / 2
        ref = np.ma.masked_invalid(ref)

        # Calculate the (n-1 - ref) difference
        diff = np.abs(inp - ref)
    elif method == "differential":
        ref = np.ma.diff(inp)

        # Find the minimum variation prior and after the n value
        diff = np.ma.zeros(inp.size, dtype=np.float64)
        diff[1:-1] = np.minimum(np.abs(ref[:-1]), np.abs(ref[1:]))

        # Make sure that only the record (n) where the difference prior and after are opposite are considered
        with np.errstate(invalid="ignore"):
            diff[1:-1][ref[:-1] * ref[1:] >= 0] = 0

        # this is necessary to make sure that the mask value is a boolean array and not the nomask value (if all values are not masked)
        diff = np.ma.masked_invalid(diff)
    else:
        msg = f'Unknown method: "{method}", only "average" and "differential" methods are available'
        raise ValueError(
            msg,
        )

    # Start with everything as passing (1)
    flag_arr = np.ma.ones(inp.size, dtype="uint8")

    # If n-1 - ref is greater than the low threshold, SUSPECT test
    if suspect_threshold:
        with np.errstate(invalid="ignore"):
            flag_arr[diff > suspect_threshold] = QartodFlags.SUSPECT

    # If n-1 - ref is greater than the high threshold, FAIL test
    if fail_threshold:
        with np.errstate(invalid="ignore"):
            flag_arr[diff > fail_threshold] = QartodFlags.FAIL

    # test is undefined for first and last values
    flag_arr[0] = QartodFlags.UNKNOWN
    flag_arr[-1] = QartodFlags.UNKNOWN

    # Check if the original data was masked
    for i in range(inp.size):
        # Check if inp is masked (original data missing)
        if inp.mask[i]:
            flag_arr[i] = QartodFlags.MISSING

        # Check if diff is masked but not in inp (this indicates that original data is not missing,
        # but the data point got masked in diff when applying the average or differential methods
        # due to trying to calculate a value using a valid value and a missing value;
        # and because of that, we are not able to apply QARTOD thus the UNKNOWN flag)
        elif diff.mask[i] and not inp.mask[i]:
            flag_arr[i] = QartodFlags.UNKNOWN

    return flag_arr.reshape(original_shape)


@add_flag_metadata(
    standard_name="rate_of_change_test_quality_flag",
    long_name="Rate of Change Test Quality Flag",
)
def rate_of_change_test(
    inp: Sequence[Real],
    tinp: Sequence[Real],
    threshold: float,
) -> np.ma.core.MaskedArray:
    """Checks the first order difference of a series of values to see if
    there are any values exceeding a threshold defined by the inputs.
    These are then marked as SUSPECT.  It is up to the test operator
    to determine an appropriate threshold value for the absolute difference not to
    exceed. Threshold is expressed as a rate in observations units per second.
    Missing and masked data is flagged as UNKNOWN.

    Parameters
    ----------
    inp
        Input data as a numeric numpy array or a list of numbers.
    tinp
        Time data as a sequence of datetime objects compatible with pandas DatetimeIndex.
        This includes numpy datetime64, python datetime objects and pandas Timestamp object.
        ie. pd.DatetimeIndex([datetime.utcnow(), np.datetime64(), pd.Timestamp.now()])
        If anything else is passed in the format is assumed to be seconds since the unix epoch.
    threshold
        A float value representing a rate of change over time, in observation units per second.

    Returns
    -------
    flag_arr
        A masked array of flag values equal in size to that of the input.

    """
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        inp = np.ma.masked_invalid(np.array(inp).astype(np.float64))

    # Save original shape
    original_shape = inp.shape
    inp = inp.flatten()

    # Start with everything as passing (1)
    flag_arr = np.ma.ones(inp.size, dtype="uint8")

    # calculate rate of change in units/second
    roc = np.ma.zeros(inp.size, dtype="float")

    tinp = mapdates(tinp).flatten()
    roc[1:] = np.abs(
        np.diff(inp) / np.diff(tinp).astype("timedelta64[s]").astype(float),
    )

    with np.errstate(invalid="ignore"):
        flag_arr[roc > threshold] = QartodFlags.SUSPECT

    # If the value is masked set the flag to MISSING
    flag_arr[inp.mask] = QartodFlags.MISSING

    return flag_arr.reshape(original_shape)


@add_flag_metadata(
    standard_name="flat_line_test_quality_flag",
    long_name="Flat Line Test Quality Flag",
)
def flat_line_test(
    inp: Sequence[Real],
    tinp: Sequence[Real],
    suspect_threshold: int,
    fail_threshold: int,
    tolerance: Real = 0,
) -> np.ma.MaskedArray:
    """Check for consecutively repeated values within a tolerance.
    Missing and masked data is flagged as UNKNOWN.
    More information: https://github.com/ioos/ioos_qc/pull/11.

    Parameters
    ----------
    inp
        Input data as a numeric numpy array or a list of numbers.
    tinp
        Time data as a sequence of datetime objects compatible with pandas DatetimeIndex.
        This includes numpy datetime64, python datetime objects and pandas Timestamp object.
        ie. pd.DatetimeIndex([datetime.utcnow(), np.datetime64(), pd.Timestamp.now()])
        If anything else is passed in the format is assumed to be seconds since the unix epoch.
    suspect_threshold
        The number of seconds within `tolerance` to allow before being flagged as SUSPECT.
    fail_threshold
        The number of seconds within `tolerance` to allow before being flagged as FAIL.
    tolerance
        The tolerance that should be exceeded between consecutive values.
        To determine if the current point `n` should be flagged, we use a rolling window, with endpoint at
        point `n`, and calculate the range of values in the window. If that range is less than `tolerance`,
        then the point is flagged.

    Returns
    -------
    flag_arr
        A masked array of flag values equal in size to that of the input.

    """
    # input as numpy arr
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        inp = np.ma.masked_invalid(np.array(inp).astype(np.float64))

    # Save original shape
    original_shape = inp.shape
    inp = inp.flatten()

    # Start with everything as passing
    flag_arr = np.full((inp.size,), QartodFlags.GOOD)

    # if we have fewer than 3 points, we can't run the test, so everything passes
    min_inp_size = 3
    if len(inp) < min_inp_size:
        return flag_arr.reshape(original_shape)

    # determine median time interval
    tinp = mapdates(tinp).flatten()

    # The thresholds are in seconds so we round make sure the interval is also in seconds
    time_interval = np.median(np.diff(tinp)).astype("timedelta64[s]").astype(float)

    def rolling_window(a, window):
        """https://rigtorp.se/2011/01/01/rolling-statistics-numpy.html."""
        if len(a) < window:
            return np.ma.MaskedArray(np.empty((0, window + 1)))
        shape = *a.shape[:-1], *(a.shape[-1] - window + 1, window + 1)
        strides = (*a.strides, a.strides[-1])
        arr = np.lib.stride_tricks.as_strided(a, shape=shape, strides=strides)
        return np.ma.masked_invalid(arr[:-1, :])

    def run_test(test_threshold, flag_value) -> None:
        # convert time thresholds to number of observations
        count = (int(test_threshold) / time_interval).astype(int)

        # calculate actual data ranges for each window
        window = rolling_window(inp, count)
        data_min = np.min(window, 1)
        data_max = np.max(window, 1)
        data_range = np.abs(data_max - data_min)

        # find data ranges that are within threshold and flag them
        test_results = np.ma.filled(data_range < tolerance, fill_value=False)
        # data points before end of first window should pass
        n_fill = min(len(inp), count)
        test_results = np.insert(test_results, 0, np.full((n_fill,), fill_value=False))
        flag_arr[test_results] = flag_value

    run_test(suspect_threshold, QartodFlags.SUSPECT)
    run_test(fail_threshold, QartodFlags.FAIL)

    # If the value is masked set the flag to MISSING
    flag_arr[inp.mask] = QartodFlags.MISSING

    return flag_arr.reshape(original_shape)


@add_flag_metadata(
    standard_name="attenuated_signal_test_quality_flag",
    long_name="Attenuated Signal Test Quality Flag",
)
def attenuated_signal_test(  # noqa: PLR0913
    inp: Sequence[Real],
    tinp: Sequence[Real],
    suspect_threshold: Real,
    fail_threshold: Real,
    test_period: Real | None = None,
    min_obs: Real | None = None,
    min_period: int | None = None,
    check_type: str = "std",
) -> np.ma.MaskedArray:
    """Check for near-flat-line conditions using a range or standard deviation.

    Missing and masked data is flagged as UNKNOWN.

    Parameters
    ----------
    inp
        Input data as a numeric numpy array or a list of numbers.
    tinp
        Time input data as a numpy array of dtype `datetime64`.
    suspect_threshold
        Any calculated value below this amount will be flagged as SUSPECT.
        In observations units.
    fail_threshold
        Any calculated values below this amount will be flagged as FAIL.
        In observations units.
    test_period
        Length of time to test over in seconds [optional].
        Otherwise, will test against entire `inp`.
    min_obs
        Minimum number of observations in window required to calculate a result [optional].
        Otherwise, test will start at beginning of time series.
        Note: you can specify either `min_obs` or `min_period`, but not both.
    min_period
        Minimum number of seconds in test_period required to calculate a result [optional].
        Otherwise, test will start at beginning of time series.
        Note: you can specify either `min_obs` or `min_period`, but not both.
    check_type
        Either 'std' (default) or 'range', depending on the type of check you wish to perform.

    Returns
    -------
    flag_arr
        A masked array of flag values equal in size to that of the input.
        This array will always contain only a single unique value since all
        input data is flagged together.

    """
    # window_func: Applied to each window when `time_period` is supplied
    # check_func: Applied to a flattened numpy array when no `time_period` is supplied
    # These are split for performance reasons
    if check_type == "std":
        window_func = lambda x: x.std()  # noqa: E731
        check_func = np.ma.std
    elif check_type == "range":

        def window_func(w):
            # When pandas>=1.0 and numba are installed, this is about twice as fast
            try:
                return w.apply(np.ptp, raw=True, engine="numba")
            except (ImportError, TypeError, NumbaTypeError):
                return w.apply(np.ptp, raw=True)

        check_func = np.ma.ptp
    else:
        msg = f'Check type "{check_type}" is not one of ["std", "range"]'
        raise ValueError(msg)

    tinp = mapdates(tinp)
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        inp = np.ma.masked_invalid(np.array(inp).astype(np.float64))

    # Save original shape
    original_shape = inp.shape

    # Start with everything as not tested (0)
    flag_arr = np.full((inp.size,), QartodFlags.UNKNOWN)

    if test_period:
        if min_obs is not None:
            min_periods = min_obs
        elif min_period is not None:
            time_interval = np.median(np.diff(tinp)).astype("timedelta64[s]").astype(float)
            min_periods = (min_period / time_interval).astype(int)
        else:
            min_periods = None
        series = pd.Series(inp.flatten(), index=tinp.flatten())
        windows = series.rolling(f"{test_period}s", min_periods=min_periods)
        check_val = window_func(windows)
    else:
        # applying np.ptp to Series causes warnings, this is a workaround
        series = inp.flatten()
        check_val = np.ones_like(flag_arr) * check_func(series)

    flag_arr[check_val >= suspect_threshold] = QartodFlags.GOOD
    flag_arr[check_val < suspect_threshold] = QartodFlags.SUSPECT
    flag_arr[np.isnan(check_val)] = QartodFlags.UNKNOWN
    flag_arr[check_val < fail_threshold] = QartodFlags.FAIL
    flag_arr[inp.mask] = QartodFlags.MISSING

    return flag_arr.reshape(original_shape)


@add_flag_metadata(
    standard_name="density_inversion_test_flag",
    long_name="Density Inversion Test Flag",
)
def density_inversion_test(
    inp: Sequence[Real],
    zinp: Sequence[Real],
    suspect_threshold: float | None = None,
    fail_threshold: float | None = None,
) -> np.ma.core.MaskedArray:
    """With few exceptions, potential water density will increase with increasing pressure. When
    vertical profile data is obtained, this test is used to flag as failed T, C, and SP observations, which
    yield densities that do not sufficiently increase with pressure. A small operator-selected density
    threshold (DT) allows for micro-turbulent exceptions. This test can be run on downcasts, upcasts,
    or down/up cast results produced in real time.

    Both Temperature and Salinity should be flagged based on the result of this test.

    Ref: Manual for Real-Time Quality Control of in-situ Temperature and Salinity Data, Version 2.0, January 2016

    Parameters
    ----------
    inp
        Potential density values as a numeric numpy array or a list of numbers.
    zinp
        Corresponding depth/pressure values for each density.
    suspect_threshold
        A float value representing a maximum potential density(or sigma0)
        variation to be tolerated, downward density variation exceeding this will be flagged as SUSPECT.
    fail_threshold
        A float value representing a maximum potential density(or sigma0)
        variation to be tolerated, downward density variation exceeding this will be flagged as FAIL.

    Returns
    -------
    flag_arr
        A masked array of flag values equal in size to that of the input.

    """
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        inp = np.ma.masked_invalid(np.array(inp).astype(np.float64))
        zinp = np.ma.masked_invalid(np.array(zinp).astype(np.float64))

    # Make sure both inputs are the same size.
    if inp.shape != zinp.shape:
        msg = f"Density ({inp.shape}) and depth ({zinp.shape}) must be the same shape"
        raise ValueError(msg)

    # Start with everything as passing
    flag_arr = QartodFlags.GOOD * np.ma.ones(inp.size, dtype="uint8")

    # If no data or just one record, return respectively an empty mask array or UNKNOWN
    if inp.size == 0:
        return np.ma.masked_array([])
    inp_size = 2
    if inp.size < inp_size:
        flag_arr[0] = QartodFlags.UNKNOWN
        return flag_arr

    # Compute the vertical density variability along zinp and flip delta according to zinp variation direction
    delta = np.sign(np.diff(zinp)) * np.diff(inp)

    if suspect_threshold is not None:
        with np.errstate(invalid="ignore"):
            is_suspect = delta < suspect_threshold
            if any(is_suspect):
                flag_arr[:-1][is_suspect == True] = QartodFlags.SUSPECT  # noqa:E712 - Previous value
                flag_arr[1:][is_suspect == True] = QartodFlags.SUSPECT  # noqa:E712 - Reversed value

    if fail_threshold is not None:
        with np.errstate(invalid="ignore"):
            is_fail = delta < fail_threshold
            if any(is_fail):
                flag_arr[:-1][is_fail == True] = QartodFlags.FAIL  # noqa:E712 - Previous value
                flag_arr[1:][is_fail == True] = QartodFlags.FAIL  # noqa:E712 - Reversed Value

    # If the value or depth is masked set the flag to MISSING for this record and the following one.
    is_missing = inp.mask | zinp.mask
    flag_arr[is_missing] = QartodFlags.MISSING
    flag_arr[1:][is_missing[:-1]] = QartodFlags.MISSING
    return flag_arr



"""Tests based on the ARGO QC manual."""

import logging
import warnings
from collections.abc import Sequence
from numbers import Real as N

import numpy as np

from ioos_qc.qartod import QartodFlags
from ioos_qc.utils import add_flag_metadata, great_circle_distance, mapdates

L = logging.getLogger(__name__)


@add_flag_metadata(
    stanard_name="pressure_increasing_test_quality_flag",
    long_name="Pressure Increasing Test Quality Flag",
)
def pressure_increasing_test(inp):
    """Returns an array of flag values where each input is flagged with SUSPECT if
    it does not monotonically increase.

    Ref: ARGO QC Manual: 8. Pressure increasing test

    Parameters
    ----------
    inp
        Pressure values as a numeric numpy array or a list of numbers.

    Returns
    -------
    flag_arr
        A masked array of flag values equal in size to that of the input.

    """
    delta = np.diff(inp)
    flags = np.ones_like(inp, dtype="uint8") * QartodFlags.GOOD

    # Correct for downcast vs upcast by flipping the sign if it's decreasing
    sign = np.sign(np.mean(delta))
    if sign < 0:
        delta = sign * delta

    flag_idx = np.where(delta <= 0)[0] + 1
    flags[flag_idx] = QartodFlags.SUSPECT

    return flags


@add_flag_metadata(
    standard_name="speed_test_quality_flag",
    long_name="Speed Test Quality Flag",
)
def speed_test(
    lon: Sequence[N],
    lat: Sequence[N],
    tinp: Sequence[N],
    suspect_threshold: float,
    fail_threshold: float,
) -> np.ma.core.MaskedArray:
    """Checks that the calculated speed between two points is within reasonable bounds.

    This test calculates a speed between subsequent points by
      * using latitude and longitude to calculate the distance between points
      * calculating the time difference between those points
      * checking if distance/time_diff exceeds the given threshold(s)

    Missing and masked data is flagged as UNKNOWN.

    If this test fails, it typically means that either a position or time is bad data,
    or that a platform is mislabeled.

    Ref: ARGO QC Manual: 5. Impossible speed test

    Parameters
    ----------
    lon
        Longitudes as a numeric numpy array or a list of numbers.
    lat
        Latitudes as a numeric numpy array or a list of numbers.
    tinp
        Time data as a sequence of datetime objects compatible with pandas DatetimeIndex.
        This includes numpy datetime64, python datetime objects and pandas Timestamp object.
        ie. pd.DatetimeIndex([datetime.utcnow(), np.datetime64(), pd.Timestamp.now()])
        If anything else is passed in the format is assumed to be seconds since the unix epoch.
    suspect_threshold
        A float value representing a speed, in meters per second.
        Speeds exceeding this will be flagged as SUSPECT.
    fail_threshold
        A float value representing a speed, in meters per second.
        Speeds exceeding this will be flagged as FAIL.

    Returns
    -------
    flag_arr
        A masked array of flag values equal in size to that of the input.

    """
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        lat = np.ma.masked_invalid(np.array(lat).astype(np.float64))
        lon = np.ma.masked_invalid(np.array(lon).astype(np.float64))
        tinp = mapdates(tinp)

    if lon.shape != lat.shape or lon.shape != tinp.shape:
        msg = f"Lon ({lon.shape}) and lat ({lat.shape}) and tinp ({tinp.shape}) must be the same shape"
        raise ValueError(msg)

    # Save original shape
    original_shape = lon.shape
    lon = lon.flatten()
    lat = lat.flatten()
    tinp = tinp.flatten()

    # If no data, return
    if lon.size == 0:
        return np.ma.masked_array([])

    # Start with everything as passing
    flag_arr = QartodFlags.GOOD * np.ma.ones(lon.size, dtype="uint8")

    # If either lon or lat are masked we just set the flag to MISSING
    mloc = lon.mask & lat.mask
    flag_arr[mloc] = QartodFlags.MISSING

    # If only one data point, return
    lon_size = 2
    if lon.size < lon_size:
        flag_arr[0] = QartodFlags.UNKNOWN
        return flag_arr.reshape(original_shape)

    # Calculate the great_distance between each point
    dist = great_circle_distance(lat, lon)

    # calculate speed in m/s
    speed = np.ma.zeros(tinp.size, dtype="float")
    speed[1:] = np.abs(
        dist[1:] / np.diff(tinp).astype("timedelta64[s]").astype(float),
    )

    with np.errstate(invalid="ignore"):
        flag_arr[speed > suspect_threshold] = QartodFlags.SUSPECT

    with np.errstate(invalid="ignore"):
        flag_arr[speed > fail_threshold] = QartodFlags.FAIL

    # first value is unknown, since we have no speed data for the first point
    flag_arr[0] = QartodFlags.UNKNOWN

    # If the value is masked set the flag to MISSING
    flag_arr[dist.mask] = QartodFlags.MISSING

    return flag_arr.reshape(original_shape)



from .config_creator import (
    QC_CONFIG_CREATOR_SCHEMA,
    VARIABLE_CONFIG_SCHEMA,
    CreatorConfig,
    QcConfigCreator,
    QcVariableConfig,
)

__all__ = [
    "QC_CONFIG_CREATOR_SCHEMA",
    "VARIABLE_CONFIG_SCHEMA",
    "CreatorConfig",
    "QcConfigCreator",
    "QcVariableConfig",
]
